import os, json
import click
import asyncio
from uvicorn import run as start_server
from . import KunYu 
from .exceptions import AnimeNotFoundError, CharacterNotFoundError


S = "\x1b[38;5;33m"
E = "\x1b[0m"
VA = "\x1b[38;5;203m"

CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".kunyu")
CONFIG_FILE = os.path.join(CONFIG_DIR, "config.json")

def create_default_config():
    """Creates the default config file in ~/.kunyu/config.json if it doesn't exist."""
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)
    if not os.path.exists(CONFIG_FILE):
        default_config = {
            "use_cache": False,
            "db_path": "cache.db",
            "port": 8000,
            "host": "127.0.0.1"
        }
        with open(CONFIG_FILE, "w") as f:
            json.dump(default_config, f, indent=4)


def load_config():
    """Loads config from ~/.kunyu/config.json."""
    if os.path.exists(CONFIG_FILE):
        with open(CONFIG_FILE, "r") as f:
            return json.load(f)
    return {}

@click.group()
def cli():
    """AnimeScraper - CLI to fetch and search anime/character details from MyAnimeList."""
    create_default_config()


@click.command()
@click.argument('anime_name')
def search_anime(anime_name: str):
    """Search for an anime by name."""
    try:
        async def run():
            async with KunYu() as scraper:
                print(f"\x1b[38;5;45m> Searching for anime\x1b[0m: \x1b[38;5;44m{anime_name}\x1b[0m\n")
                anime = await scraper.search_anime(anime_name)

                click.echo(f"{S}Title{E}: {VA}{anime.title}{E}")
                click.echo(f"{S}Japanese Title{E}: {VA}{anime.japanese_title}{E}")
                click.echo(f"{S}Type{E}: {VA}{anime.anime_type}{E}")
                click.echo(f"{S}Episodes{E}: {VA}{anime.episodes}{E}")
                click.echo(f"{S}Duration{E}: {VA}{anime.duration}")
                click.echo(f"{S}Status{E}: {VA}{anime.status}{E}")
                click.echo(f"{S}Rating{E}: {VA}{anime.rating}{E}")
                click.echo(f"{S}Genres{E}: {VA}{' '.join(anime.genres)}{E}")
                click.echo(f"{S}Themes{E}: {VA}{' '.join(anime.themes)}{E}")
                click.echo(f"{S}Premiered{E}: {VA}{anime.premiered}{E}")
                click.echo(f"{S}Aired{E}: {VA}{anime.aired}{E}")
                click.echo(f"{S}Score{E}: {VA}{anime.stats.score}{E}")
                click.echo(f"{S}Scored by{E}: {VA}{anime.stats.scored_by} people{E}")
                click.echo(f"{S}Popularity{E}: {VA}{anime.stats.popularity}{E}")
                click.echo(f"{S}Ranked{E}: {VA}{anime.stats.ranked}{E}")
                click.echo(f"{S}Synopsis{E}: \x1b[38;5;35m{anime.synopsis[:300]}...\x1b[0m")

        asyncio.run(run())
    except AnimeNotFoundError as e:
        click.echo(f"Error: {e}", err=True)



@click.command()
@click.argument('anime_id')
def get_anime(anime_id: str):
    """Get anime details by anime ID."""
    try:
        async def run():
            async with KunYu() as scraper:
                print(f"\x1b[38;5;45m> Fetching anime details for ID: \x1b[0m: \x1b[38;5;44m{anime_id}\x1b[0m")

                anime = await scraper.get_anime(anime_id)

                genres = " ".join(anime.genres)
                click.echo(f"{S}Title{E}: {VA}{anime.title}{E}")
                click.echo(f"{S}Japanese Title{E}: {VA}{anime.japanese_title}{E}")
                click.echo(f"{S}Type{E}: {VA}{anime.anime_type}{E}")
                click.echo(f"{S}Episodes{E}: {VA}{anime.episodes}{E}")
                click.echo(f"{S}Duration{E}: {VA}{anime.duration}")
                click.echo(f"{S}Status{E}: {VA}{anime.status}{E}")
                click.echo(f"{S}Rating{E}: {VA}{anime.rating}{E}")
                click.echo(f"{S}Genres{E}: {VA}{genres}{E}")
                click.echo(f"{S}Themes{E}: {VA}{' '.join(anime.themes)}{E}")
                click.echo(f"{S}Premiered{E}: {VA}{anime.premiered}{E}")
                click.echo(f"{S}Aired{E}: {VA}{anime.aired}{E}")
                click.echo(f"{S}Score{E}: {VA}{anime.stats.score}{E}")
                click.echo(f"{S}Scored by{E}: {VA}{anime.stats.scored_by} people{E}")
                click.echo(f"{S}Popularity{E}: {VA}{anime.stats.popularity}{E}")
                click.echo(f"{S}Ranked{E}: {VA}{anime.stats.ranked}{E}")
                click.echo(f"{S}Synopsis{E}: \x1b[38;5;35m{anime.synopsis[:300]}...\x1b[0m")
        asyncio.run(run())

    except AnimeNotFoundError as e:
        click.echo(f"Error: {e}", err=True)



@click.command()
@click.argument('character_id')
def get_character(character_id: str):
    """Get character details by character ID."""
    try:
        async def run():
            async with KunYu() as scraper:

                print(f"\x1b[38;5;45m> Fetching character details for ID: \x1b[0m: \x1b[38;5;44m{character_id}\x1b[0m\n")
                character = await scraper.get_character(character_id)

                click.echo(f"{S}Name{E}: {VA}{character.name}{E}")
                click.echo(f"{S}Japanese Name{E}: {VA}{character.japanese_name}")
                for key, value in character.about.items():
                    click.echo(f"{S}{key}{E}: {VA}{value}{E}")
                click.echo(f"{S}Description{E}: \x1b[38;5;35m{character.description[:300]}...{E}")  # Limit the description

        asyncio.run(run())
    except CharacterNotFoundError as e:
        click.echo(f"Eror: {e}")



@click.command()
@click.argument('character_name')
def search_character(character_name: str):
    """Get character details by character name."""
    try: 
        async def run():
            async with KunYu() as scraper:

                print(f"\x1b[38;5;45m> Searching for character\x1b[0m: \x1b[38;5;44m{character_name}\x1b[0m\n")
                character = await scraper.search_character(character_name)

                click.echo(f"{S}Name{E}: {VA}{character.name}{E}")
                click.echo(f"{S}Japanese Name{E}: {VA}{character.japanese_name}")
                for key, value in character.about.items():
                    click.echo(f"{S}{key}{E}: {VA}{value}{E}")
                click.echo(f"{S}Description{E}: \x1b[38;5;35m{character.description[:300]}...{E}")  # Limit the description
        asyncio.run(run())

    except CharacterNotFoundError as e:
        click.echo(f"Eror: {e}")


@click.command()
@click.option("--host", default=None, help="Host for the API server (overrides config.json)")
@click.option("--port", default=None, help="Port for the API server (overrides config.json)")
@click.option("--use-cache", is_flag=True, help="Enable database caching (overrides config.json)")
@click.option("--db-path", default=None, help="Path for the local SQLite cache database (overrides config.json)")
def server(host: str, port: int, use_cache: bool, db_path: str):
    """Start the FastAPI server for AnimeScraper."""
    
    # Load from config file and merge with CLI args
    config = load_config()

    # Override config values with CLI arguments if provided
    final_host = host if host else config.get("host", "127.0.0.1")
    final_port = port if port else config.get("port", 8000)
    final_use_cache = use_cache if use_cache else config.get("use_cache", False)
    final_db_path = db_path if db_path else config.get("db_path", "cache.db")

    click.echo(f"🚀 Starting server on http://{final_host}:{final_port}")
    click.echo(f"📁 Database Path: {final_db_path} | 📦 Use Cache: {final_use_cache}")
    # Pass the user arguments to the server through environment variables
    os.environ["ANIME_SCRAPER_USE_CACHE"] = str(final_use_cache)
    os.environ["ANIME_SCRAPER_DB_PATH"] = final_db_path

    # Run the FastAPI server
    start_server("AnimeScraper.animescraper_server:app", host=final_host, port=int(final_port), reload=True)

        
# Add all commands to the CLI
cli.add_command(search_anime)
cli.add_command(search_character)
cli.add_command(get_anime)
cli.add_command(get_character)
cli.add_command(server)

if __name__ == '__main__':
    cli()






"""
Data models for AnimeScraper.

This module defines data structures for anime, characters, and related entities.
"""
from typing import Optional, List, Dict
#from dataclasses import dataclass 
from pydantic import BaseModel


class AnimeCharacter(BaseModel):
    """
    Represents a character in an anime.

    Attributes:
        id (str): The character's ID.
        name (str): The character's name.
        role (Optional[str]): The character's role in the anime.
        voice_actor (Dict[str, str]): Details about the voice actor.
    """
    id: str
    name: str
    role: Optional[str]
    voice_actor: Dict[str, str]

    # def _to_dict(self):
    #     return self.__dict__
    #
    # @classmethod
    # def _from_dict(cls, data: dict):
    #     return cls(**data)
    #

class Character(BaseModel):
    """
    Represents a character with full details.

    Attributes:
        id (str): The character's ID.
        name (str): The character's name.
        japanese_name (str): The character's Japanese name.
        about (Dict): A dictionary containing The character's Age, Height, Weight etc.
        description (str): The character's description from MAL.
        img (str): Character image.
        favorites (str): The number people marked it as favorite
        url (str): The MAL Page url of the Character.
    """
    id: str
    name: str
    japanese_name: Optional[str]
    about: Dict[str, str]
    description: str
    img: str
    favorites: str
    url: str

    # def _to_dict(self):
    #     return self.__dict__
    #
    # @classmethod
    # def _from_dict(cls, data: dict):
    #     return cls(**data)

class AnimeStats(BaseModel):
    """
    Represents The stats of an Anime.

    Attributes:
        score (str): The Anime's score.
        scored_by (str): scored by the number of poeple.
        ranked (str): The rank of The Anime.
        popularity (str): The popularity of the Anime.
        favorites (stt): The number of poeple marked it as their favorite Anime.
    """

    score: str
    scored_by: str
    ranked: str
    popularity: str
    members: str
    favorites: str

    # def _to_dict(self):
    #     return self.__dict__
    #
    # @classmethod
    # def _from_dict(cls, data: dict):
    #     return cls(**data)
    #

class Anime(BaseModel):
    """
    Represents an anime.

    Attributes:
        id (str): The anime's ID.
        title (str): The anime's title.
        english_title (Optional[str]): The English title of the anime.
        japanese_title (Optional[str]): The Japanese title of the anime.
        anime_type (str): The type of anime (e.g., TV, Movie).
        episodes (Optional[str]): Number of episodes.
        status (str): The current status (e.g., Finished Airing).
        aired (str): The airing date range.
        duration (str): Average duration of an episode.
        rating (str): The age rating (e.g., PG-13).
        synopsis (str): A brief synopsis of the anime.
        genres (List[str]): List of genres.
        studios (str): List of studios.
        themes (List[str]): List of themes.
        stats (AnimeStats): Statistics about the anime.
        characters (List[AnimeCharacter]): List of characters in the anime.
    """

    id: str
    title: str
    english_title: Optional[str]
    japanese_title: Optional[str]
    anime_type: str    # TV, Movie, OVA etc.
    episodes: Optional[str]
    status: str       # Airing, Finished, etc.
    aired: str
    duration: str
    premiered: str
    rating: str # PG-13, R, etc.
    synopsis: str
    genres: List[str]
    studios: str
    themes: List[str]
    stats: AnimeStats
    characters: List[AnimeCharacter]


    # def _to_dict(self):
    #     return {
    #         **self.__dict__,
    #         "stats": self.stats.__dict__,
    #         "characters": [c._to_dict() for c in self.characters]
    #     }
    #
    # @classmethod
    # def _from_dict(cls, data: dict):
    #     stats = AnimeStats._from_dict(data["stats"])
    #     characters = [AnimeCharacter._from_dict(char) for char in data["characters"]]
    #     return cls(
    #         id=data["id"],
    #         title=data["title"],
    #         english_title=data["english_title"],
    #         japanese_title=data["japanese_title"],
    #         anime_type=data["anime_type"],
    #         episodes=data["episodes"],
    #         status=data["status"],
    #         aired=data["aired"],
    #         duration=data["duration"],
    #         premiered=data["premiered"],
    #         rating=data["rating"],
    #         synopsis=data["synopsis"],
    #         genres=data["genres"],
    #         studios=data["studios"],
    #         themes=data["themes"],
    #         stats=stats,
    #         characters=characters,
    #     )
    #



"""
This is the Synchronus version of `KunYu()` class.

This module contains the **SyncKunYu()** class, which acts as the primary entry point
for interacting with `Myanimelist <https://myanimelist.net/>`__ data, as well as helper classes and functions.
"""

from typing import Optional, List
import httpx
from ._model import Anime, Character
from .sync_malscraper import SyncMalScraper

class SyncKunYu:

    """
    The main Class for interacting with MAL Synchronusly.
    """
    def __init__(
        self, 
        use_cache: bool = False,
        db_path: str = "cache.db",
        timeout: int = 10
    ) -> None:

        """
        If you want to cache the data locally pass ``use_cache=True`` and specify the database path where you want to cache ``db_path='cache.db'.`` It uses `sqlite3` for storing data.

        Args:
            use_cache (bool): If data should be cached. (Default: False)
            db_path: (str): The path of the database. (Default: cache.db)
        """


        self.shared_client: Optional[httpx.Client] = None
        self.use_cache = use_cache
        self.db_path = db_path
        self.timeout = timeout
        self.Scraper = SyncMalScraper(
            client=self.shared_client,
            use_cache=self.use_cache,
            db_path=self.db_path,
            timeout=self.timeout,
        )
    

    def __enter__(self):
        self.shared_client = httpx.Client(headers={
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1"

        })
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.shared_client:
            self.shared_client.close()

    def search_anime(self, anime_name: str)-> Anime:
        """
        Fetches and Returns Anime details from myanimelist.

        Args:
            anime_name (str): Name of the anime you want to search.
        
        Returns:
            Anime: Returns Anime object with anime details.

        Example:
            >>> from AnimeScraper import SyncKunYu
            >>> scraper = SyncKunYu()
            >>> anime = scraper.search_anime("dr stone")
            >>> print(anime.title)

        """

        with self.Scraper as scraper:
            anime = scraper.search_anime(anime_name)
            return anime


    def search_character(self, character_name: str)-> Character:
        """
        Fetches and Returns Character details from myanimelist.

        Args:
            character_name (str): Name of the Character you want to search.
        
        Returns:
            Character: Returns Character object with the character details.

        Example:
            >>> from AnimeScraper import SyncKunYu
            >>> with SyncKunYu() as scraper:
            >>>     character = scraper.search_character("Togashi Yuuta")
            >>>     print(character.name)

        Notes:
            You can use ``with SyncKunYu`` context manager for same session use.
        """
        with self.Scraper as scraper:
            character = scraper.search_character(character_name)
            return character 


    def get_anime(self, anime_id: str)->Anime:
        """
        Fetches anime details from MyAnimeList.

        Args:
            anime_id (str): The ID of the anime.

        Returns:
            Anime: An object containing anime details.
        """
        with self.Scraper as scraper:
            anime = scraper.get_anime(anime_id)
            return anime

    def get_character(self, character_id: str)-> Character:
        """
        Fetches character details from MyAnimeList.

        Args:
            character_id (str): The ID of the character.

        Returns:
            Character: An object containing character details.
        """
        with self.Scraper as scraper:
            character = scraper.get_character(character_id)
            return character

    def search_batch_anime(self, anime_names: List[str])-> List[Anime]:
        """
        Fetches anime details in batch.
        Args:
            anime_names (List(str)): List of anime names.

        Returns:
            List[Anime]: A list of Anime objects with Anime details.
        """

        with self.Scraper as scraper:
            anime_list = scraper.search_batch_anime(anime_names)

        return anime_list
    


    def search_batch_character(self, character_names: List[str])-> List[Character]:
        """
        Fetches character details in batch.
        Args:
            character_names (List(str)): List of characters name.

        Returns:
            List[Character]: A list of Character objects with character's details.
        """

        with self.Scraper as scraper:
            batch_characters = scraper.search_batch_character(character_names)

        return batch_characters
